class Minsky:
    """Simulateur de machines à compteurs."""

    def __init__(self, instructions=None, état_initial=None, programme=None):
        """Initialise le simulateur, soit en donnant directement 
        les `instructions et l’`état_initial`, soit en 
        compilant le `programme`."""
        if programme is not None:
            self.compile(programme)
        elif instructions is not None:
            self.instructions = instructions
            self.état = état_initial
            self.calcule_largeur_affichage()

    def calcule_largeur_affichage(self):
        # utilisé uniquement pour l’affichage avec __str__() ci-dessous
        self.largeur = max(*(len(état) for état in self.instructions))

    def compile(self, programme):
        """Compile un `programme` pour une machine à compteur,
        et retourne la machine.

        Une instruction est une ligne de la forme :

            état_courant +compteur état_suivant

        ou :

            état_courant -compteur état_si_nul état_si_décrémente

        Les espaces et l’indentation sont libres.
        Toute ligne non conforme est ignorée silencieusement."""

        self.instructions = {}
        self.état = None

        lignes = programme.split('\n')
        for ligne in lignes:
            instruction = ligne.split()
            if len(instruction) == 3:
                compteur = instruction[1]
                if len(compteur) >= 2 and compteur[0] == '+':
                    compteur = compteur[1:]
                    self.ajoute_incr(instruction[0], compteur, instruction[2])
            if len(instruction) == 4:
                compteur = instruction[1]
                if len(compteur) >= 2 and compteur[0] == '-':
                    compteur = compteur[1:]
                    self.ajoute_test(instruction[0], compteur, instruction[2],
                            instruction[3])

        self.calcule_largeur_affichage()
        return self


    def ajoute_incr(self, courant, compteur, suivant):
        """Rajoute une instruction d’incrémentation.
        
        Si c’est la première instruction, c’est l’état initial."""
        
        if self.état is None:
            self.état = courant
        self.instructions[courant] = compteur, suivant


    def ajoute_test(self, courant, compteur, suivant_si_nul, suivant_si_decr):
        """Rajoute une instruction de test et décrémentation.
        
        Si c’est la première instruction, c’est l’état initial."""
        
        if self.état is None:
            self.état = courant
        self.instructions[courant] = compteur, suivant_si_decr, suivant_si_nul


    def calcule_un_pas(self):
        """Effectue une étape de la machine suivant son instructions."""
        instruction = self.instructions[self.état]
        if len(instruction) == 2:
            id_compteur, self.état = instruction
            self.compteurs[id_compteur] += 1
        else:
            id_compteur, état_1, état_0 = instruction
            compteur = self.compteurs[id_compteur]
            if compteur > 0:
                self.compteurs[id_compteur] = compteur - 1
                self.état = état_1
            else:
                self.état = état_0

    
    def calcule(self, compteurs, trace=False):
        """Fait tourner la machine: le dictionnaire `compteurs`
        donne les valeurs initiales.
        
        Si `trace` est mis à `True`, on affiche les étapes successives."""

        self.compteurs = compteurs

        if trace: print(self)
        while self.état in self.instructions:
            self.calcule_un_pas()
            if trace: print(self)

    def __str__(self):
        compteurs = ', '.join(f"{nom}: {valeur}"
                for nom, valeur in self.compteurs.items())
        return f"{self.état:{self.largeur}}: {compteurs}"

if __name__ == "__main__":
    compteurs = {'A': 3, 'B': 4, 'C': 0}
    machine = Minsky().compile("""
		A_vers_C     -A  B_vers_C A_vers_C_bis
		A_vers_C_bis +C  A_vers_C
		B_vers_C     -B  fini B_vers_C_bis
		B_vers_C_bis +C  B_vers_C
    """)
    print(machine.instructions)
    machine.calcule(compteurs,trace=True)
