class Ruban:
    """Ruban de machine de Turing."""

    def __init__(self, caractères="", décalage=0, blanc="."):
        """Initialise un ruban.

        Par défaut, le ruban est vierge. On peut spécifier le caractère
        spécial `blanc` qui représente les cases non écrites (on utilise un
        point par défaut). On peut également fournir des `caractères`
        initialement écrits sur le ruban, l’argument `décalage` donnant l’écart
        entre la tête et le *dernier* caractère fourni (un décalage positif
        fait revenir la tête vers le début de la chaîne)."""

        self._blanc = blanc

        # La moitié gauche du ruban, qui commence au caractère courant
        # (indices positifs), est représentée par une liste (les cases au-delà
        # de la longueur de la liste sont vides).
        self._partie_gauche = list(caractères)

        # La moitié droite du ruban, qui commence après le caractère courant
        # (indices strictement négatifs), est représentée par une liste
        # (l’indice -1 est le dernier élement de la liste, ce qui permet
        # d’utiliser pop() et append()).
        self._partie_droite = []

        self.bouge(décalage)

    def gauche(self):
        """Décale la tête du ruban vers la gauche."""
        try:
            self._partie_droite.append(self._partie_gauche.pop())
        except IndexError:
            self._partie_droite.append(self._blanc)

    def droite(self):
        """Décale la tête du ruban vers la droite."""
        try:
            self._partie_gauche.append(self._partie_droite.pop())
        except IndexError:
            self._partie_gauche.append(self._blanc)

    def bouge(self, décalage):
        """Déplace la tête du nombre de cases indiquées par le `décalage`.
        
        Un décalage positif déplace vers la gauche."""

        if décalage > 0:
            décale = self.gauche
        else:
            décale = self.droite
            décalage = -décalage
        for _ in range(décalage):
            décale()

    def écrit(self, caractère):
        """Écrit un `caractère` à l’emplacement courant de la tête de lecture."""
        try:
            self._partie_gauche[-1] = caractère
        except IndexError:
            self._partie_gauche = [caractère]

    def lit(self):
        """Lit le caractère courant."""
        try:
            return self._partie_gauche[-1]
        except IndexError:
            return self._blanc

    def __repr__(self):
        caractères = "".join(self._partie_gauche) + "".join(reversed(self._partie_droite))
        décalage   = len(self._partie_droite)
        return f'<Ruban(blanc="{self._blanc}",caractères="{caractères}",décalage="{décalage}")>'

    def __str__(self, sep=" "):
        """Renvoie la chaîne des caractères écrits sur le ruban, séparés par
        des espaces (ou la chaîne donnée comme `sep`), le caractère courant
        étant encadré par les signes `<` et `>`."""
        try:
            à_gauche = sep.join(self._partie_gauche[:-1])
            tête     = self._partie_gauche[-1]
        except IndexError:
            à_gauche = ""
            tête     = self._blanc
        à_droite = sep.join(reversed(self._partie_droite))
        return f"{à_gauche}<{tête}>{à_droite}"


class Turing:
    """Simulateur de machines de Turing à plusieurs rubans."""

    def __init__(self, delta, état_initial, liste_rubans):
        self.état         = état_initial
        self.delta        = delta
        self.rubans       = liste_rubans

    def calcule_un_pas(self):
        """Effectue une étape de la machine suivant sa fonction de transition."""
        état, caractères, mouvements = self.delta[self.état][tuple(ruban.lit() for ruban in self.rubans)]
        self.état = état
        for numéro in range(len(self.rubans)):
            ruban = rubans[numéro]
            ruban.écrit(caractères[numéro])
            ruban.bouge(mouvements[numéro])

    def calcule(self,trace=False):
        """Fait tourner la machine jusqu’à obtenir un résultat.
        
        Si `trace` est mis à `True`, on affiche les étapes successives."""

        if trace: print(self)
        while self.état in self.delta:
            self.calcule_un_pas()
            if trace: print(self)

    def __str__(self):
        rubans = " / ".join([str(ruban) for ruban in self.rubans])
        return f"{self.état}: {rubans}"


if __name__ == "__main__":
    palindrome = {
            'I': {
                # on recopie le premier ruban sur le second, en sens inverse
                ('0','.'): ('I',('0','0'),(1,-1)),
                ('1','.'): ('I',('1','1'),(1,-1)),
                ('.','.'): ('R',('.','.'),(-1,0)),
                },
            'R': {
                # on revient au début du premier ruban
                ('0','.'): ('R',('0','.'),(-1,0)),
                ('1','.'): ('R',('1','.'),(-1,0)),
                ('.','.'): ('V',('.','.'),(1,1)),
                },
            'V': {
                # on vérifie si ça colle
                ('0','0'): ('V',('0','0'),(1,1)),
                ('1','1'): ('V',('1','1'),(1,1)),
                ('.','.'): ('O',('.','.'),(0,0)),
                ('0','1'): ('N',('0','1'),(0,0)),
                ('1','0'): ('N',('1','0'),(0,0)),
                }
            }
    rubans = [Ruban("...00110...",décalage=3),Ruban("........",décalage=6)]
    machine = Turing(palindrome, 'I', rubans).calcule(trace=True)


